package JTwilight;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;
import java.util.Vector;
import java.util.Map.Entry;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Lock;


// Twilight STM's runtime system. Shared variables must be represented by
// JTwilight.GlobalVar. Then, transactional operations with twilight code can be
// specified by passing an instance of JTwilight.TransactionBody to the
// static method atomic.
public class Transaction<R> {
	// Global clock, restart counter and zone-access-lock
	private static AtomicInteger _T = new AtomicInteger(0);
	private static ZoneLock _L = new ZoneLock();

	private HashMap<GlobalVar<?>, ReadEntry> _readset;
	private HashMap<GlobalVar<?>, Stack<WriteEntry>> _writeset;

	// Set of locks acquired by the transaction. These locks are released
	// when the transaction restarts or finishes.
	private Set<Lock> _locks;

	// transaction versions/time stamps (t_start and t_commit)
	private int _transactionStartVersion;
	private int _commitVersion;

	// some state flags
	private boolean _isInTwilightZone;
	private boolean _alreadyDidPreAcquire;
    private boolean _readsetInconsistent; // keeps track of the consistency of the entire read set

	// parent transaction
	private Transaction<?> _parent;

	// Initialization
	private Transaction(Transaction<?> ancestor) {
		_parent = ancestor;
		if (ancestor == null) {
			_readset = new HashMap<GlobalVar<?>, ReadEntry>();
			_writeset = new HashMap<GlobalVar<?>, Stack<WriteEntry>>();
		} else {
			_readset = ancestor._readset;
			_writeset = ancestor._writeset;
            
            if (ancestor._isInTwilightZone) {
                throw new IllegalStateException("Cannot create a nested transaction in the twilight zone.");
            }
		}

		_locks = new HashSet<Lock>();

		_transactionStartVersion = _T.get(); 
		_commitVersion = _transactionStartVersion;

		_isInTwilightZone = false;
		_alreadyDidPreAcquire = false;
        _readsetInconsistent = false;
	}

	private HashMap<GlobalVar<?>, WriteEntry> privateWriteSet() {
		HashMap<GlobalVar<?>, WriteEntry> privateWrites = new HashMap<GlobalVar<?>, WriteEntry>();
		for( Entry<GlobalVar<?>, Stack<WriteEntry>> e : _writeset.entrySet()) {
			Stack<WriteEntry>  wstack = e.getValue();
			WriteEntry we = wstack.peek();
			if (we._writer == this) {
				privateWrites.put(e.getKey(), wstack.peek());
			}
		}
		return privateWrites;
	}

	// Runs the given transaction with twilight zone
	public static <R> R execute(TransactionBody<R> body, Transaction<?> parent) {
		// repeat until the transaction finishes or is aborted
		while (true) {
			Transaction<R> trans = new Transaction<R>(parent);

			trans.validateReadset();
			Set<GlobalVar<?>> inconsistencies = trans.inconsistencies();
			// if there are inconsistencies, outer transactions need to restart
			if (!inconsistencies.isEmpty()) {
				Iterator<GlobalVar<?>> iterator = inconsistencies.iterator();
				ReadEntry re = trans._readset.get(iterator.next());
				Transaction<?> bad = re._readers.peek();
				parent.restart(bad);
			}

			body.stm = trans;

			// move the start time of the transaction forward
			for (Transaction<?> ancestor = trans._parent; ancestor != null; ancestor = ancestor._parent) {
				ancestor._transactionStartVersion = trans._transactionStartVersion;
			}
			
			try {
				body.transactional();
				trans.prepare();

				trans._isInTwilightZone = true;
				R result = body.twilight();

				trans.finish();

				// move the time frame
				for (Transaction<?> ancestor = trans._parent; ancestor != null; ancestor = ancestor._parent) {
					ancestor._transactionStartVersion = trans._commitVersion;
				}
				return result;
			} catch (RestartException restart) {
				// only continue when reached the appropriate parent.
				// otherwise: throw e; to re-throw to the one parent up.

				// clean up state of read and write sets first
				Iterator<ReadEntry> iter = trans._readset.values().iterator();
				while (iter.hasNext()) {
					ReadEntry re = iter.next();
					if (re._readers.peek() == trans) {
						re._readers.pop();
						if (re._readers.isEmpty())
							iter.remove();
					}
				}
				
				Iterator<Stack<WriteEntry>> witer = trans._writeset.values().iterator();
				while (witer.hasNext()) {
					Stack<WriteEntry>  wstack = witer.next();
					WriteEntry we = wstack.peek();
					if (we._writer == trans) {
						wstack.pop();
						if (wstack.empty())
							witer.remove();
					}
				}
				
				if (restart.badTransaction == trans) {
					continue;
				} else {
					trans._parent.restart(restart.badTransaction);
				}
				
			} finally {
				// cleanup
				trans.releaseLocks();

                // inessential cleanup (but removes the possibility of accidental copying of
                // the transaction in the TransactionBody)
                trans = null;
                body.stm = null;
			}
		}
	}

	// locks the given set of variables in a fixed order. Adds the locks to
	// _locks,
	// such that the locks are released when the transaction finishes or
	// restarts.
	private void acquireExclusiveAccess(Set<GlobalVar<?>> vars) {
		for (GlobalVar<?> v : new TreeSet<GlobalVar<?>>(vars)) {
			v._lock.lock();
			_locks.add(v._lock);
		}
	}
	
	private boolean tryAcquireExclusiveAccess(Set<GlobalVar<?>> vars) {
		for (GlobalVar<?> v : new TreeSet<GlobalVar<?>>(vars)) {
				if (!v._lock.tryLock()) {
					releaseLocks();
					return false;
				}
			_locks.add(v._lock);
		}
		return true;
	}

	// release the locks in _locks.
	private void releaseLocks() {
		for (Lock l : _locks)
			l.unlock();
	}

	// prepares the transaction for entering the twilight zone and committing.
	private void prepare() {
		boolean succeeded = false;
		while(!succeeded) {
			_L.accessPrepare();
			succeeded = tryAcquireExclusiveAccess(_writeset.keySet());
			_L.releasePrepare();
			
			if (!succeeded)
				Thread.yield();
		}

		_commitVersion = _T.incrementAndGet();
		validateReadset();
	}

	// checks if the read set is coherent with respect to the start time of the
	// transaction.
	private void validateReadset() {
		for (Map.Entry<GlobalVar<?>, ReadEntry> e : _readset.entrySet()) {
			GlobalVar<? extends Object> var = e.getKey();
			ReadEntry entry = e.getValue();

			if (_writeset.containsKey(var)) {
				// we've got this variable locked
				entry._inConflict = var._version > _transactionStartVersion;
			} else {
				// make sure that the variable was not locked at one moment in
				// time during the resampling of the clock and reading the
				// version.
				boolean b = var._lock.isLocked();
				entry._inConflict = b || var._version > _transactionStartVersion;
			}
            
            // An entry in conflict makes the entire readset inconsistent.
            _readsetInconsistent = _readsetInconsistent || entry._inConflict;

			// these two flags are still the same up till this point. Later in
			// the twilight zone entry._inConflict only implies
			// entry._inConflictSinceStart.
			entry._inConflictSinceStart = entry._inConflict;
		}
	}

	// finish the transaction if the read set is consistent.
	@SuppressWarnings("unchecked")
	private void finish() {
        if (_readsetInconsistent)
            restart(this);

		Iterator<Entry<GlobalVar<?>, Stack<WriteEntry>>> iterator = _writeset.entrySet().iterator();
		while (iterator.hasNext()) {
			Entry<GlobalVar<?>, Stack<WriteEntry>> e = iterator.next();
			GlobalVar var = e.getKey();
			Stack<WriteEntry> wstack = e.getValue();
            
            //using open nesting semantics, the write gets published to shared global state
			if (wstack.peek()._writer == this) {
				WriteEntry we = wstack.pop();
				var._version = _commitVersion;
				var._value = we._value;
                
				//publish the write to the parents sets
				if (wstack.empty()) {
					iterator.remove();
				} else {
					wstack.peek()._value = we._value;
				}

				// publish the writes to the parents read set
				ReadEntry re = _readset.get(var);
				if (re != null) {
					re._value = we._value;
					re._version = _commitVersion;
				}
			}
			
		}

		Iterator<ReadEntry> iter = _readset.values().iterator();
		while (iter.hasNext()) {
			ReadEntry re = iter.next();
			if (re._readers.peek() == this) {
				re._readers.pop();
				if (re._readers.isEmpty())
					iter.remove();
			}
		}
	}

	// restarts the transaction. We use the exception mechanism for this.
	public void restart(Transaction<?> bad) {
		throw new RestartException(bad);
										
	}

	// extends the read set. May be used in the transactional zone too.
	// returns true if extending is done in a consistent way, false otherwise.
	public <T> boolean extendReadset(GlobalVar<T> var) {

		if (_readset.containsKey(var)) {
			ReadEntry re = _readset.get(var);
			if (re._readers.peek() != this) {
				re._readers.push(this);
			}
			return true; // read set already contains this variable.
		}

		// non-blocking read of the value. Version needs to be read twice to
		// detect concurrent modifications.
		int v1 = var._version;
		boolean l1 = var._lock.isLocked();
		T val = var._value;
		boolean l2 = var._lock.isLocked();
		int v2 = var._version;

		boolean consistentVersion = v1 == v2 && !l1 && !l2 && v1 == v2;

		ReadEntry entry = new ReadEntry();
		entry._readers = new Stack<Transaction<?>>();
		entry._readers.push(this);
		entry._version = v1;
		entry._value = val;
		entry._inConflict = !consistentVersion || v1 > _commitVersion;
		entry._inConflictSinceStart = !consistentVersion
				|| v1 > _transactionStartVersion;
        
         // any entry inconssistent: then the entire read set inconsistent
        _readsetInconsistent = _readsetInconsistent || entry._inConflict;

		_readset.put(var, entry);

		return entry._inConflict;
	}

	// transactionally read from a shared variable.
	@SuppressWarnings("unchecked")
	public <T> T read(GlobalVar<T> var) {
		if (var == null)
			throw new IllegalArgumentException(
					"Tried to read the non-initialized global var " + var);

		// return the value in the transactions write set if it's there.
		// write set of parent must not get searched
		Stack<WriteEntry> wstack = _writeset.get(var);
		if (wstack != null) {
			if (wstack.peek()._writer == this) {
				return (T) wstack.peek()._value;
			}

		}

		// read set has to be extended manually in the application in the
		// twilight zone.
		if (!_isInTwilightZone) {
			extendReadset(var);
		}
		
		ReadEntry entry = _readset.get(var);
		if (entry == null) {
			throw new IllegalStateException("Tried to read " + var
					+ " which was not in the read set");
		}
		
		if (entry._inConflict) {
			restart(this);
		}
		
		return (T) entry._value;
	}

	// transactionally write to a shared variable.
	// writes are done lazily and only recorded to be published later.
	public <T> void write(GlobalVar<T> var, T value) {
		if (var == null) {
			throw new IllegalArgumentException(
					"Tried to write to a null pointer " + var);
		}
		
		if (_isInTwilightZone && !_writeset.containsKey(var)) {
			throw new IllegalStateException("Tried to write " + var
					+ " which was not in the write set");
		}
		
		Stack<WriteEntry> wstack = _writeset.get(var);
		// first write of this var
		if (wstack == null) {
			wstack = new Stack<WriteEntry>();
			WriteEntry we = new WriteEntry();	
			we._writer = this;
			we._value = value;
			wstack.push(we);
		} else {
			//var has been written to before
			if (wstack.peek()._writer == this) {
				wstack.peek()._value = value;
			} else {
				WriteEntry we = new WriteEntry();	
				we._writer = this;
				we._value = value;
				wstack.push(we);
			}
		}
		
		_writeset.put(var, wstack);
			
	}

	// temporarily drops the locks on the write set's variables.
	// uses the locking protocol.
	public void preAcquire() {
		if (_alreadyDidPreAcquire)
			return;

		_L.accessPreacquire();
		releaseLocks();

		_alreadyDidPreAcquire = true;
	}

	// atomically rereads the read set, and moves the transaction forward in
	// time. Temporarily drops the locks on the write set and regains them.
	// for inner transactions, atomize can result in an implicit restart.
	public void atomize() {
		// drop the locks, and prevent other transactions from entering the
		// twilight zone.
		preAcquire();

        Map<GlobalVar<?>,WriteEntry> privateWrites = privateWriteSet();
        if (_readsetInconsistent) {
            // lock read set + write set in predetermined order.
            Set<GlobalVar<?>> union = new HashSet<GlobalVar<?>>();
            union.addAll(privateWrites.keySet());
            union.addAll(_readset.keySet());
            acquireExclusiveAccess(union);
        }
        else {
            // lock only the write set
            acquireExclusiveAccess(privateWrites.keySet());
        }

		// allow other transactions to enter the twilight zone again.
		_L.releasePreacquire();

		_commitVersion = _T.incrementAndGet();
        
        if (_readsetInconsistent) {
            // reread the current values. Atomically because we hold the locks.
            for (Map.Entry<GlobalVar<?>, ReadEntry> e : _readset.entrySet()) {
                GlobalVar<?> var = e.getKey();
                ReadEntry entry = e.getValue();
    
                entry._inConflictSinceStart = var._version > _transactionStartVersion;
                
                if (entry._readers.size() > 1 || entry._readers.peek() != this) {
                    this.restart(entry._readers.peek()); 
                }
                
                entry._version = var._version;
                entry._value = var._value;
                entry._inConflict = false;
                
                // update write-set entry if the variable was also in the writeset
                Stack<WriteEntry> wstack = _writeset.get(var);
                if (wstack != null) {
                    if (wstack.peek()._writer == this) {
                        wstack.peek()._value = var._value;
                    }
                }
    
                if (!privateWrites.containsKey(var)) {
                    var._lock.unlock();
                    _locks.remove(var._lock);
                }
            }
        }
	}

	// Set containing all the inconsistencies since the start of the
	// transaction.
	public Set<GlobalVar<?>> inconsistencies() {
        if (!_readsetInconsistent) {
            return new HashSet<GlobalVar<?>>(0); // optimization: if there are no inconsistencies, then no need to iterate through all the read set entries
        }
        
		HashSet<GlobalVar<?>> result = new HashSet<GlobalVar<?>>();
		for (Map.Entry<GlobalVar<?>, ReadEntry> e : _readset.entrySet()) {
			GlobalVar<?> var = e.getKey();
			ReadEntry entry = e.getValue();

			if (entry._inConflictSinceStart)
				result.add(var);
		}

		return result;
	}

	// read set entry.
	private static class ReadEntry {
		Stack<Transaction<?>> _readers;
		int _version;
		Object _value;
		boolean _inConflict;
		boolean _inConflictSinceStart;
	}

	private static class WriteEntry {
		Transaction<?> _writer;
		Object _value;
	}

	

	// transactional code and twilight code should be wrapped inside a
	// TransactionBody. It offers a "stm" field which can be used to access
	// the stm operations.
	// Note that both methods to be implemented are re-executed when the
	// transaction restarts.
	public static abstract class TransactionBody<R> {
		protected Transaction<R> stm;

		public TransactionBody() {
			stm = null;
		}

		abstract public void transactional();

		abstract public R twilight();
	}

	// multiple reader/single writer lock that is biased towards writers;
	// the releasing thread does not have to be the thread who acquired the
	// lock.
	// also: method names are changed to more appropriate names concerning the
	// function.
	private static class ZoneLock {
		//Counter == 0 : you can enter prepare or preacquire
		//Counter < 0  : some transaction is in prepare
		//Counter > 0  : some transaction is in preacquire
		private int _counter;
		//private int _accessors;
		//private int _blockers;
		private int _requestPreacquire;

		public ZoneLock() {
			_counter = 0;
			_requestPreacquire = 0;
		}

		public synchronized void accessPrepare() {
			while (_counter > 0 || _requestPreacquire > 0) {
				try {
					wait();
				} catch (InterruptedException e) {
				}
			}

			--_counter;
		}

		public synchronized void releasePrepare() {
			++_counter;
			notifyAll();
		}

		public synchronized void accessPreacquire() {
			++ _requestPreacquire;
			while (_counter < 0) {
				try {
					wait();
				} catch (InterruptedException e) {
				}
			}

			++_counter;
		}

		public synchronized void releasePreacquire() {
			--_counter;
			--_requestPreacquire;
			notifyAll();
		}
	}

	

	// Some test example.
	public static void main(String[] args) {

		// two shared variables
		final GlobalVar<Number> x = new GlobalVar<Number>(42);
		final GlobalVar<Integer> y = new GlobalVar<Integer>(10);
		final GlobalVar<String> z = new GlobalVar<String>("Blub");

		Transaction.execute(new Transaction.TransactionBody<Void>() {
			Number v;
			Integer w;

			public void transactional() {
				String blub = stm.read(z);
				v = stm.read(x);
				stm.write(x, v);
				w = stm.read(y);
				stm.write(x, w);
				stm.write(z, blub.toUpperCase());
			}

			public Void twilight() {
				System.out.println(w);
				return null;
			}
		}, null);

		// read-only transaction
		Transaction.execute(new Transaction.TransactionBody<Void>() {
			public void transactional() {
				stm.read(x);
				stm.read(z);
			}

			public Void twilight() {
				System.out.println("z has become: " + stm.read(z));
				return null;
			}
		}, null);

		// restarting transaction
		Transaction.execute(new Transaction.TransactionBody<Void>() {
			boolean test = true;
			String s = "";

			public void transactional() {
				if (test) {
					test = false;
					s += "First try failed.";
					stm.restart(stm);
				} else {
					s += "But the next one worked.";
				}
			}

			public Void twilight() {
				System.out.println(s);
				return null;
			}
		}, null);
		
		final GlobalVar<Vector<Integer>> vec = new GlobalVar<Vector<Integer>>(new Vector<Integer>());
		// pointers.... careful!!!
		Transaction.execute(new Transaction.TransactionBody<Void>() {
			
			public void transactional() {
				Vector<Integer> localVec = stm.read(vec);
				localVec.add(1);
			}

			public Void twilight() {
				System.out.println(stm.read(vec).isEmpty());
				return null;
			}
		}, null);

		// nested transaction
		Transaction.execute(new Transaction.TransactionBody<Void>() {
			int a;

			public void transactional() {
				a = stm.read(y);
				stm.write(z, "Blah");

				Transaction.execute(
						new Transaction.TransactionBody<Void>() {
							String b;

							public void transactional() {
								stm.write(y, a + 1);
								b = stm.read(z);
							}

							public Void twilight() {
								System.out.println("z was changed by parent, but I wasn't told: "
										+ b);
								return null;
							}
						}, stm);
			}

			public Void twilight() {
				System.out.println("y has become: " + stm.read(y));
				System.out.println("z has become: " + stm.read(z));
				
				return null;
			}
		}, null);
	}
}
