/* =============================================================================
 *
 * stm.h
 *
 * User program interface for STM. For an STM to interface with STAMP, it needs
 * to have its own stm.h for which it redefines the macros appropriately.
 *
 * =============================================================================
 *
 * Author: Chi Cao Minh
 *
 * =============================================================================
 */


#ifndef STM_H
#define STM_H 1


#include "tl2.h"
#include "util.h"

#define STM_THREAD_T                    Thread
#define STM_SELF                        Self
#define STM_RO_FLAG                     ROFlag

#define STM_MALLOC(size)                TxAlloc(STM_SELF, size)
#define STM_FREE(ptr)                   TxFree(STM_SELF, ptr)


#  define malloc(size)                  tmalloc_reserve(size)
#  define calloc(n, size)               ({ \
                                            size_t numByte = (n) * (size); \
                                            void* ptr = tmalloc_reserve(numByte); \
                                            if (ptr) { \
                                                memset(ptr, 0, numByte); \
                                            } \
                                            ptr; \
                                        })
#  define realloc(ptr, size)            tmalloc_reserveAgain(ptr, size)
#  define free(ptr)                     tmalloc_release(ptr)

#  include <setjmp.h>
#  define STM_JMPBUF_T                  sigjmp_buf
#  define STM_JMPBUF                    buf


#define STM_VALID()                     (1)
#define STM_RESTART()                   TxAbort(STM_SELF)

#define STM_STARTUP()                   TxOnce()
#define STM_SHUTDOWN()                  TxShutdown()

#define STM_NEW_THREAD()                TxNewThread()
#define STM_INIT_THREAD(t, id)          TxInitThread(t, id)
#define STM_FREE_THREAD(t)              TxFreeThread(t)








#  define STM_BEGIN(isReadOnly)         do { \
                                            STM_JMPBUF_T STM_JMPBUF; \
                                            int STM_RO_FLAG = isReadOnly; \
                                            sigsetjmp(STM_JMPBUF, 1); \
                                            TxStart(STM_SELF, &STM_JMPBUF, &STM_RO_FLAG); \
                                        } while (0) /* enforce comma */

#define STM_BEGIN_RD()                  STM_BEGIN(1)
#define STM_BEGIN_WR()                  STM_BEGIN(0)
#define STM_END()                       { if (TxBeginCommit(STM_SELF) == 0) { TxEndCommit(STM_SELF); } else { TxAbort(STM_SELF); } }
#define STM_INITIATE_END()              TxBeginCommit(STM_SELF)
#define STM_FINISH_END()                TxEndCommit(STM_SELF)

typedef volatile intptr_t               vintp;

#define STM_READ(var)                   TxLoad(STM_SELF, (vintp*)(void*)&(var), 0)
#define STM_READ_F(var)                 IP2F(TxLoad(STM_SELF, \
                                                    (vintp*)FP2IPP(&(var)), 0))
#define STM_READ_P(var)                 IP2VP(TxLoad(STM_SELF, \
                                                     (vintp*)(void*)&(var), 0))

#define STM_WRITE(var, val)             TxStore(STM_SELF, \
                                                (vintp*)(void*)&(var), \
                                                (intptr_t)(val), 0)
#define STM_WRITE_F(var, val)           TxStore(STM_SELF, \
                                                (vintp*)FP2IPP(&(var)), \
                                                F2IP(val), 0)
#define STM_WRITE_P(var, val)           TxStore(STM_SELF, \
                                                (vintp*)(void*)&(var), \
                                                VP2IP(val), 0)

#define STM_WRITE_UPDATE(var, val)      TxStoreUpdate(STM_SELF, \
                                                (vintp*)(void*)&(var), \
                                                (intptr_t)(val))

#define STM_WRITE_WITH_HISTORY(var, val) TxStore(STM_SELF, \
                                                (vintp*)(void*)&(var), \
                                                (intptr_t)(val), 1)
#define STM_READ_WITH_HISTORY(h)         TxLoad(STM_SELF, (vintp*)(void*)&(h), 1)

#define STM_LOCAL_WRITE(var, val)       ({var = val; var;})
#define STM_LOCAL_WRITE_F(var, val)     ({var = val; var;})
#define STM_LOCAL_WRITE_P(var, val)     ({var = val; var;})

#define STM_READ_FROM_READSET(var)            GetReadsetValue(STM_SELF, (vintp*) (void*) &(var))
#define STM_MARK_INCOHERENT(var)              MarkIncoherent(STM_SELF, (vintp*) (void*) &(var))
#define STM_EXTEND_READSET(var)               ExtendReadset(STM_SELF, (vintp*) (void*) &(var), 0)
#define STM_EXTEND_READSET_WITH_HISTORY(var)  ExtendReadset(STM_SELF, (vintp*) (void*) &(var), 1)
#define STM_REMOVE(var)                       RemoveFromReadset(STM_SELF, (vintp*) (void*) &(var))

#define STM_RELOAD_READSET()                  while(LoadReadsetWithLocks(STM_SELF)) { usleep(0); printf("Locking the readset failed. Retrying.\n"); }
#define STM_NEXT_INCOHERENCE(prev,cur)        NextIncoherence(STM_SELF, (prev), (cur))

#define STM_REMOVE_ZONE(number)               ForgetZone(STM_SELF, (number))
#define STM_ENTER_ZONE(number)                EnterZone(STM_SELF, (number))
#define STM_LEAVE_ZONE()                      LeaveZone(STM_SELF)


#endif /* STM_H */


/* =============================================================================
 *
 * End of stm.h
 *
 * =============================================================================
 */
